import pytest
from unittest.mock import patch, mock_open
from gat_llm.tools.make_custom_plot import ToolMakeCustomPlot


@pytest.fixture
def mock_exec():
    with patch("builtins.exec") as mock:
        yield mock


@pytest.fixture
def mock_os_path_isfile():
    with patch("os.path.isfile") as mock:
        mock.return_value = True
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    tmcp = ToolMakeCustomPlot()
    ans = tmcp("import matplotlib.pyplot as plt", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_make_custom_plot_success(mock_exec, mock_os_path_isfile):
    tmcp = ToolMakeCustomPlot()
    plot_code = """
import matplotlib.pyplot as plt
plt.plot([1, 2, 3, 4])
plt.ylabel('some numbers')
plt.savefig('media/plot.jpg')
"""
    result = tmcp(plot_code)
    assert "<image>" in result
    assert "<path_to_image>" in result
    assert "</image>" in result


def test_make_custom_plot_execution_error(mock_exec):
    mock_exec.side_effect = Exception("Execution error")
    tmcp = ToolMakeCustomPlot()
    plot_code = "invalid_code"
    result = tmcp(plot_code)
    assert "Plot was NOT generated" in result
    assert "Execution error" in result


@patch("os.path.isfile", return_value=False)
def test_make_custom_plot_file_not_saved(mock_isfile, mock_exec):
    tmcp = ToolMakeCustomPlot()
    plot_code = """
import matplotlib.pyplot as plt
plt.plot([1, 2, 3, 4])
plt.ylabel('some numbers')
plt.savefig('media/plot.jpg')
"""
    result = tmcp(plot_code)
    assert "Error: Image was not saved correctly" in result


# Add more tests for different scenarios and edge cases



import re
import requests
from urllib.parse import urljoin
from bs4 import BeautifulSoup, NavigableString, Tag

from bs4.element import Comment


def extract_visible_html(html):
    soup = BeautifulSoup(html, "html.parser")

    # Remove script, style, meta, head, link, hidden, aria-hidden, and other non-visible elements
    for hidden in soup(
        ["script", "style", "meta", "head", "link", "[hidden]", '[aria-hidden="true"]']
    ):
        hidden.decompose()

    # Remove elements with inline styles that hide them
    for element in soup.find_all(True):
        style = element.get("style", "")

        # Check for display: none, visibility: hidden, or opacity: 0 in the style attribute
        if (
            "display: none" in style
            or "visibility: hidden" in style
            or "opacity: 0" in style
        ):
            element.decompose()

    # Remove comment nodes
    for comment in soup.findAll(string=lambda text: isinstance(text, Comment)):
        comment.extract()

    # Loop through all elements and remove tags that are empty or hidden
    for element in soup.find_all(True):
        if not element.get_text(strip=True):  # Remove elements without visible text
            element.decompose()

    # Return the modified HTML as a string
    return soup.prettify()


def get_text_and_urls(url_content):
    # Parse the HTML content
    soup = BeautifulSoup(url_content, "html.parser")

    # Find all <a> tags and replace them with formatted URLs
    for link in soup.find_all("a"):
        href = link.get("href")
        text = link.get_text()
        new_tag = soup.new_tag("a", href=href)
        new_tag.string = text
        link.replace_with(new_tag)

    # Extract the modified HTML content
    output = str(soup)
    return output


def tag_visible(element):
    if element.parent.name in ["a", "p"]:
        return True

    if element.parent.name in [
        "style",
        "script",
        "head",
        "title",
        "meta",
        # "[document]",
    ]:
        return False
    if isinstance(element, Comment):
        return False
    return True


def text_from_html(body):
    soup = MyBeautifulSoup(body, "html.parser")
    texts = soup.findAll(string=True)
    visible_texts = filter(tag_visible, texts)

    url_list = []
    for link in soup.find_all("a"):
        url_list.append(f"[ {link.get('href')} ] {link.text}")
        # print(link.get('href'))

    return " ".join(t.strip() for t in visible_texts), "\n".join(url_list)


# https://stackoverflow.com/questions/52026274/beautiful-soup-get-all-text-but-preserve-link-html
class MyBeautifulSoup(BeautifulSoup):
    def get_text(self, separator="", strip=False, types=(NavigableString,)):
        text_parts = []

        for element in self.descendants:
            if isinstance(element, NavigableString):
                text_parts.append(str(element))
            elif isinstance(element, Tag):
                if element.name == "a" and "href" in element.attrs:
                    text_parts.append(
                        element.get_text(separator=separator, strip=strip)
                    )
                    text_parts.append("( " + element["href"] + " )")
                elif isinstance(element, types):
                    text_parts.append(
                        element.get_text(separator=separator, strip=strip)
                    )

        return separator.join(text_parts)


class ToolGetUrlContent:
    def __init__(self, query_llm, max_subpages_to_read=60):
        self.name = "get_url_content"
        self.query_llm = query_llm
        self.max_subpages_to_read = max_subpages_to_read

        self.tool_description = {
            "name": self.name,
            "description": f"""Retrieves the contents of one or more internet URLs specified by internet_urls, separated by commas. For example, use this tool when the user requests you to read a webpage, or find something on the internet. A prompt should be provided to extract only the desired relevant content from the URLs. If the prompt is left empty, the raw contents of the webpage are returned. Note that the prompt will be provided individually to query each URL.

You can use {self.name} in these <use_cases></use_cases> and others as needed when:
<use_cases>
<use_case>It is easy to retrieve up-to-date information from a webpage</use_case>
<use_case>It is necessary to navigate internet web pages to retrieve information</use_case>
<use_case>The user wants to browse websites containing news. In this case, follow the steps:

- Read the initial URL using recursion_level=0 to find the URLs to the webpages linked.
- Find a pattern that will retrieve only the relevant URLs. In general, strings like "/news/" or "/economy/" will be contained in the URL.
- Read the initial URL using recursion_level=1 and the appropriate recursion_regex_condition to retrieve only relevant news.
</use_case>
</use_cases>

Raises ValueError: if the request is invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "internet_urls": {
                        "type": "string",
                        "description": """Website URLs separated by commas, as in the <url_examples></url_examples>:
<url_examples>
<url_example>http://www.folha.com</url_example>
<url_example>http://www.g1.com, https://www.walljournal.com, http://websitename.com.uk</url_example>
</url_examples>""",
                    },
                    "prompt": {
                        "type": "string",
                        "description": """Query prompt to retrieve specific information from a single web page. For example:
<website_prompt_examples>
<website_prompt_example>Summarize the contents of this web page. Make sure to include all links in the page.</website_prompt_example>
<website_prompt_example>Read this web page and extract only the URLs related to economic news.</website_prompt_example>
</website_prompt_examples>
Leave argument prompt empty ("") to retrieve the raw text.
""",
                    },
                    "return_all_visible_html": {
                        "type": "string",
                        "enum": ["True", "False"],
                        "description": """Allowed choices are True and False.
If True, returns all raw HTML that contains visible elements.
If False, only returns the visible text of the website and a list of URLs found.""",
                    },
                    "recursion_level": {
                        "type": "integer",
                        "default": 0,
                        "description": """Read all links from the original webpage recursion_level times. Default is 0.
If recursion_level is greater than 0, also read and return the contents of links found in the webpages.
This is useful to automatically follow links and retrieve detailed contents from pages.
For example: if the user requests reading news from a webpage, setting recursion_level to 1 will usually also retrieve the contents of the articles.
""",
                    },
                    "recursion_regex_condition": {
                        "type": "string",
                        "description": """If recursion_level is greater than 0, child pages must match this regex to be loaded.
Use this parameter to avoid reading an excessive amount of useless data in subpaged.
Note that the webpage URL will be checked using python's regex re package using: re.match(recursion_regex_condition, URL)
""",
                    },
                },
                "required": ["internet_urls"],
            },
        }

    def __call__(
        self,
        internet_urls,
        prompt="",
        return_all_visible_html=False,
        recursion_level=0,
        recursion_regex_condition="",
        **kwargs,
    ):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        return_all_visible_html = str(return_all_visible_html).lower().strip() == "true"
        recursion_level = int(recursion_level)
        internet_urls = internet_urls.split(",")
        internet_urls = [x.strip() for x in internet_urls]
        ans = []
        for u in internet_urls:
            ans.append("<url_content>")
            ans.append(f"<url>{u}</url>")
            content = self._get_url_content(
                u,
                return_all_visible_html,
                max_recursion_level=recursion_level,
                recursion_regex_condition=recursion_regex_condition,
            )
            if prompt.strip() != "":
                if self.query_llm is None:
                    content = "Could not answer prompt because a Language Model was not provided."
                else:
                    sys_prompt = "Read the contents of the following webpage to answer questions:"
                    sys_prompt = (
                        sys_prompt + f"\n<webpage_contents>{content}<webpage_contents>"
                    )
                    sys_prompt = (
                        sys_prompt + f"\nAlways include relevant links in your answers."
                    )
                    llm_ans = self.query_llm(
                        prompt,
                        system_prompt=sys_prompt,
                    )
                    for x in llm_ans:
                        pass
                    content = x

            ans.append(f"<content>{content}</content>")
            ans.append("</url_content>")
        return "\n".join(ans)

    def _get_url_content(
        self,
        internet_url,
        return_all_visible_html,
        cur_recursion_level=0,
        max_recursion_level=0,
        recursion_regex_condition="",
    ):
        """Retrieves URL contents

        Arguments:
            internet_url: URL to read
            return_all_visible_html: whether to return all HTML that renders visible elements or just the text
            cur_recursion_level: current recursion level, for cases when subpages need to be retrieved
            max_recursion_level: maximum recursion level to go to
        """
        # keep track of pages already visited to avoid looping
        if cur_recursion_level == 0:
            self.retrieved_pages = {internet_url: True}

        try:
            headers = {
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"
            }
            c = requests.get(internet_url, headers=headers)

            if str(return_all_visible_html).lower().strip() == "true":
                # return all visible HTML (remove only scripts and hidden elements)
                visible_html = extract_visible_html(c.content)
                ans = f"<source_url>{c.url}</source_url><status_code>{c.status_code}</status_code>\n<contents>{visible_html}</contents>"
            else:
                # only extract texts and URLs
                texts, urls = text_from_html(c.text)
                # don't return URLs if navigating sub-URLs
                if max_recursion_level == 0:
                    ans = f"<source_url>{c.url}</source_url><status_code>{c.status_code}</status_code>\n<contents>{texts}</contents><urls>{urls}</urls>"
                else:
                    ans = f"<source_url>{c.url}</source_url><status_code>{c.status_code}</status_code>\n<contents>{texts}</contents>"

            # if the user requested sublinks:
            if cur_recursion_level < max_recursion_level:
                sub_links = self._extract_links(c.content)
                sub_links = [
                    x
                    for x in sub_links
                    if (
                        re.match(recursion_regex_condition, x)
                        or recursion_regex_condition in x
                    )
                ]
                assert (
                    len(sub_links) < self.max_subpages_to_read
                ), f"""Error: tried to read too many sublinks: {len(sub_links)}.
The maximum allowed is {self.max_subpages_to_read}.
Please narrow down the number of webpages using a recursion_regex_condition.
Tried to read the following pages:
{sub_links}
"""
                sub_contents = []
                for sub_link in sub_links:
                    if not self.retrieved_pages.get(sub_link, False):
                        sub_content = self._get_url_content(
                            sub_link,
                            return_all_visible_html,
                            cur_recursion_level + 1,
                            max_recursion_level,
                        )
                        self.retrieved_pages[sub_link] = True
                        sub_contents.append(sub_content)
                ans = ans + "\n" + "\n".join(sub_contents)

            return ans
        except Exception as e:
            return f"Could not retrieve page from URL {internet_url}.\nError description: {str(e)}"

        return str(ans)

    def _extract_links(self, html_text, base_url=None):
        """
        Extracts all URLs from the given HTML text.

        Parameters:
            html_text (str): The HTML content to parse.
            base_url (str): The base URL to resolve relative links (optional).

        Returns:
            list: A list of all extracted URLs.
        """
        soup = BeautifulSoup(html_text, "html.parser")
        links = []

        # Find all 'a' tags with an 'href' attribute
        for a_tag in soup.find_all("a", href=True):
            href = a_tag["href"]
            if base_url:
                # Resolve relative links against the base URL
                href = urljoin(base_url, href)
            links.append(href)

        return [x.strip() for x in links]



import pytest
from gat_llm.tools.solve_python_code import ToolSolvePythonCode


def test_unexpected_arg(unexpected_param_msg):
    tspc = ToolSolvePythonCode()
    ans = tspc("print('Hello, World!')", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_solve_python_code_success():
    tspc = ToolSolvePythonCode()
    python_code = """
text = "Hello, World!"
ans = text[::-1]
"""
    result = tspc(python_code)
    assert result == "!dlroW ,olleH"


def test_solve_python_code_execution_error():
    tspc = ToolSolvePythonCode()
    python_code = "invalid_code"
    result = tspc(python_code)
    assert "Code did NOT execute correctly" in result
    assert "Error description" in result


@pytest.mark.parametrize(
    "code,expected",
    [
        (
            "ans = sorted([3, 1, 4, 1, 5, 9, 2, 6, 5, 3])",
            "[1, 1, 2, 3, 3, 4, 5, 5, 6, 9]",
        ),
        ("ans = ''.join(sorted(set('mississippi')))", "imps"),
        ("ans = sum([x for x in range(1, 101) if x % 2 == 0])", "2550"),
        ("ans = {x: x**2 for x in range(1, 6)}", "{1: 1, 2: 4, 3: 9, 4: 16, 5: 25}"),
    ],
)
def test_solve_python_code_various_operations(code, expected):
    tspc = ToolSolvePythonCode()
    result = tspc(code)
    assert result == expected


# Add more tests for different scenarios and edge cases



import os


class ToolUpdateUserDetails:
    def __init__(self):
        self.requires_username = True
        self.name = "read_write_user_details"

        self.tool_description = {
            "name": self.name,
            "description": """Reads or writes relevant information about the user being assisted. Always READ before you WRITE. Before saving new data, show all the changes (before/after) to the user and ask for confirmation. Never save or replace information without first confirming with the user that the changes are correct. Use [] instead of <> when creating groups.

Always retrieve the existing information first so you don't forget to save important existing data. When saving, make sure to include all information still relevant, as shown below in  <user_example></user_example>:

<user_example>
If the current user information is:

[user_info]
[name]Julian[/name]
[friends]Karlo, Joseph, Milly[/friends]
[/user_info]

And the user says: Joseph is not my friend. After confirmation, the user information should be updated to:

[user_info]
[name]Julian[/name]
[friends]Karlo, Milly[/friends]
[/user_info]

All previous relevant information has to be sent in the contents.
</user_example>""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "action": {
                        "type": "string",
                        "enum": ["READ", "WRITE"],
                        "description": """One of READ or WRITE, depending whether information about the user should be retrieved or written. Always READ before using WRITE to make sure important information isn't discarded.
Before using WRITE, make sure to show to the user the changes that will be made and that they agree with them""",
                    },
                    "contents": {
                        "type": "string",
                        "description": "If action=WRITE, content that will fully replace the relevant information about the user. Do not forget to write previous data that is still relevant",
                    },
                },
                "required": ["action"],
            },
        }

    def __call__(self, action, username="None", contents="", **kwargs):
        """
        debug = [action, username, contents]
        with open("debug.txt", "w") as f:
            f.write(str(debug))
        """

        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        actions = ["READ", "WRITE"]
        if action not in actions:
            return f"Action must be one of {actions}"

        os.makedirs("user_info", exist_ok=True)
        filename = f"user_info/{username}.txt"

        if action == "READ":
            if os.path.isfile(filename):
                with open(filename, "r", encoding="utf-8") as f:
                    ans = f.read()
                return ans
            else:
                return "No information available."
        elif action == "WRITE":
            with open(filename, "w", encoding="utf-8") as f:
                f.write(contents)
            return "User information saved successfully."



import types
from unittest.mock import Mock

from gat_llm.llm_interface import LLMInterface


def test_return_any_answer():
    llm = Mock(return_value=["<scratchpad>Thoughts</scratchpad>Bot response"])
    llm.tool_use_added_msgs = []
    llm.word_counts = [2]
    rpg = None
    llm_tools = None

    li = LLMInterface("You are a helpful assistant", llm, llm_tools, rpg)
    response_ui = li.chat_with_function_caller(
        "Hello", None, ui_history=[], username=""
    )
    assert isinstance(
        response_ui, types.GeneratorType
    ), "UI Response has to be a generator"

    for x in response_ui:
        pass

    _, scratchpad_info, _, cur_history = x

    assert scratchpad_info == "Thoughts", "Scratchpad info not identified"
    assert cur_history[-1] == {
        "role": "assistant",
        "content": "Bot response\n",
    }, "Unexpected bot response"



import pytest
from gat_llm.tools.solve_symbolic import ToolSolveSymbolic


def test_unexpected_arg(unexpected_param_msg):
    tss = ToolSolveSymbolic()
    ans = tss("import sympy", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_solve_symbolic_success():
    tss = ToolSolveSymbolic()
    sympy_code = """
import sympy as sp
x = sp.Symbol('x')
expr = x**2 + 2*x + 1
ans = sp.solve(expr)
"""
    result = tss(sympy_code)
    assert result == "[-1]"


def test_solve_symbolic_execution_error():
    tss = ToolSolveSymbolic()
    sympy_code = "invalid_code"
    result = tss(sympy_code)
    assert "Code did NOT execute correctly" in result
    assert "Error description" in result


@pytest.mark.parametrize(
    "code,expected",
    [
        (
            "import sympy as sp\nx = sp.Symbol('x')\nans = sp.integrate(sp.sin(x), x)",
            "-cos(x)",
        ),
        (
            "import sympy as sp\nx, y = sp.symbols('x y')\nans = sp.diff(x**2 + y**2, x)",
            "2*x",
        ),
        (
            "import sympy as sp\nx = sp.Symbol('x')\nans = sp.limit(sp.sin(x)/x, x, 0)",
            "1",
        ),
    ],
)
def test_solve_symbolic_various_operations(code, expected):
    tss = ToolSolveSymbolic()
    result = tss(code)
    assert result == expected


# Add more tests for different scenarios and edge cases



import re
import json
import time

from openai import OpenAI
from .base_service import LLM_Service


class LLM_GPT_OpenAI(LLM_Service):
    def __init__(self, model_size):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        self.openai_client = None
        if model_size == "GPT4o OpenAI":
            self.model_id = "gpt-4o-2024-05-13"
            self.llm_description = (
                "OpenAI GPT4o (Large-size LLM) - directly from OpenAI"
            )
            self.price_per_M_input_tokens = 5
            self.price_per_M_output_tokens = 15
        elif model_size == "GPT3_5 OpenAI":
            self.model_id = "gpt-3.5-turbo-0125"
            self.llm_description = (
                "OpenAI 3.5 Turbo (Medium-size LLM) - directly from OpenAI"
            )
            self.price_per_M_input_tokens = 0.5
            self.price_per_M_output_tokens = 1.5
        elif model_size == "GPT4o mini OpenAI":
            self.model_id = "gpt-4o-mini-2024-07-18"
            self.llm_description = (
                "OpenAI 4o Mini (Small-size LLM) - directly from OpenAI"
            )
            self.price_per_M_input_tokens = 0.15
            self.price_per_M_output_tokens = 0.6

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 4000,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        In the case of OpenAI, the input is already the expected format
        """
        ans = msg_list
        return ans

    def _prepare_call_list_from_history(
        self, system_prompt, msg, b64image, chat_history
    ):
        """Prepares the prompt for the next interaction with the LLM"""
        history_list = [
            {"role": "system", "content": system_prompt},
        ]
        for x in chat_history:
            if isinstance(x, dict):
                history_list.append(x)
            else:
                history_list.append({"role": "user", "content": x[0]})
                history_list.append({"role": "assistant", "content": str(x[1])})

        if b64image is None:
            history_list.append({"role": "user", "content": msg})
        else:
            history_list.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{b64image}",
                            },
                        },
                        {"type": "text", "text": msg},
                    ],
                }
            )
        return history_list

    def invoke_streaming(
        self,
        prompt,
        b64image=None,
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=25,
    ):
        """
        Invokes the OpenAI model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered.
            In the case of Claude 3, it is a dictionary with keys
            system -> str, messages -> list
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        tools: description of the tools that can be used, Claude format
        tool_invoker_fn: function that invokes the tools. Arguments are:
                function name - function to call
                return_results_only - we set to True because we already use Claude format
                kwargs - arguments to the tool that will be called
        :return: Inference response from the model.
        """
        # create client if it hasn't been created already
        if self.openai_client is None:
            self.openai_client = OpenAI()

        # Messages that had to be added because of function use
        self.tool_use_added_msgs = []

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        if len(extra_stop_sequences) > 0:
            body["stop"] = extra_stop_sequences
        else:
            body.pop("stop", None)
        body["messages"] = prompt

        if tools is None:
            body["messages"].append({"role": "assistant", "content": postpend})
        else:
            adj_tools = []
            for x in tools:
                cur_desc = x.copy()
                cur_desc["parameters"] = cur_desc["input_schema"]
                cur_desc.pop("input_schema", None)
                adj_tools.append(
                    {
                        "type": "function",
                        "function": cur_desc,
                    }
                )
            body["tools"] = adj_tools
            # assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

        cur_fail_sleep = 60
        for k in range(max_retries):
            try:
                self.debug_body = body
                llm_body_changed = True
                while llm_body_changed:
                    llm_body_changed = False
                    response = self.openai_client.chat.completions.create(**body)
                    word_count = len(re.findall(r"\w+", str(body["messages"])))
                    print(f"Invoking {self.llm_description}. Word count: {word_count}")

                    # stream responses
                    partial_ans = self._response_gen(response, postpend)
                    x = ""
                    for x in partial_ans:
                        yield x
                    cur_ans = x

                    if self.cur_tool_spec is not None:
                        # tool use has been required. Let's do it
                        # TODO: update upstream to reflect the inclusion of a response
                        # TODO: probably rework gradio UI to re-instantiate things every chat, or keep an instance per chat ID
                        tool_ans = tool_invoker_fn(
                            self.cur_tool_spec["tool_name"],
                            return_results_only=True,
                            **self.cur_tool_spec["input"],
                        )

                        # append assistant responses
                        assistant_msg = {
                            "role": "assistant",
                            "content": [
                                {
                                    "type": "text",
                                    "text": cur_ans,
                                },
                            ],
                            "tool_calls": [
                                {
                                    "id": self.cur_tool_spec["id"],
                                    "type": "function",
                                    "function": {
                                        "name": self.cur_tool_spec["tool_name"],
                                        "arguments": json.dumps(
                                            self.cur_tool_spec["input"]
                                        ),
                                    },
                                }
                            ],
                        }

                        next_user_msg = {
                            "role": "tool",
                            "content": tool_ans,
                            "tool_call_id": self.cur_tool_spec["id"],
                        }
                        body["messages"].append(assistant_msg)
                        body["messages"].append(next_user_msg)

                        # keep a log of messages that had to be appended due to tool use
                        self.tool_use_added_msgs.append(assistant_msg)
                        self.tool_use_added_msgs.append(next_user_msg)
                        llm_body_changed = True

                    # TODO: Include proper token count and pricing
                    ans_word_count = len(
                        re.findall(r"\w+", postpend + cur_ans + str(self.stop_reason))
                    )
                    self.word_counts.append(
                        {
                            "request_word_count": word_count,
                            "answer_word_count": ans_word_count,
                            "price_estimate": 0.001
                            * (0.003 * word_count + 0.015 * ans_word_count),
                            "exec_time_in_s": time.time() - t0,
                        }
                    )
                return
            except Exception as e:
                print(
                    f'Error {str(e)}. Prompt length: {len(str(body["messages"]))}\n\nRetrying {k}...'
                )
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2

        raise

    def _response_gen(self, response_body, postpend=""):
        cur_ans = ""
        cur_tool_spec = None
        for x in response_body:
            txt = ""
            if x.choices[0].delta.tool_calls is not None and cur_tool_spec is None:
                cur_tool_spec = x.choices[0].delta.tool_calls[0].__dict__.copy()
                cur_tool_spec["arguments"] = ""

            txt = (
                x.choices[0].delta.content
                if x.choices[0].delta.content is not None
                else ""
            )
            if x.choices[0].delta.tool_calls is not None:
                cur_tool_spec["arguments"] += (
                    x.choices[0].delta.tool_calls[0].function.arguments
                )

            if txt != "":
                cur_ans += txt
                yield postpend + cur_ans

            stop_reason = x.choices[0].finish_reason
            if stop_reason is not None and stop_reason == "stop_sequence":
                stop_txt = x.delta.stop_sequence
                yield postpend + cur_ans + stop_txt
                break
        if cur_tool_spec is not None:
            cur_tool_spec["arguments"] = cur_tool_spec["arguments"].split("{")[1:]
            cur_tool_spec["arguments"] = "{" + "{".join(cur_tool_spec["arguments"])

            # print(f'*{cur_tool_spec["arguments"]}*')
            cur_tool_spec["input"] = (
                cur_tool_spec["arguments"]
                if isinstance(cur_tool_spec["arguments"], dict)
                else json.loads(cur_tool_spec["arguments"])
            )
            cur_tool_spec["tool_name"] = cur_tool_spec["function"].name
            cur_tool_spec.pop("index", None)
            cur_tool_spec.pop("type", None)
            cur_tool_spec.pop("function", None)
            cur_tool_spec.pop("arguments", None)

        self.cur_tool_spec = cur_tool_spec
        self.stop_reason = stop_reason



import re
import copy
import json
import time

from .base_service import LLM_Service


class LLM_Command_Cohere(LLM_Service):
    # TODO: Native function calling
    def __init__(self, bedrock_client, model_size):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        self.bedrock_client = bedrock_client

        if model_size == "Command R Cohere 1":
            self.model_id = "cohere.command-r-v1:0"
            self.llm_description = (
                "Cohere Command R 1.0 (Medium-size LLM) - from Bedrock"
            )
            self.price_per_M_input_tokens = 0.5
            self.price_per_M_output_tokens = 1.5
        elif model_size == "Command RPlus Cohere 1":
            self.model_id = "cohere.command-r-plus-v1:0"
            self.llm_description = (
                "Cohere Command R+ 1.0 (Large-size LLM) - from Bedrock"
            )
            self.price_per_M_input_tokens = 3
            self.price_per_M_output_tokens = 15

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 3900,
            "temperature": 0.5,  # 0.5 is default,
            "force_single_step": False,
            # "top_k": 250,
            # "top_p": 1,
            "stop_sequences": [],  # the regular is already implemented
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        ans = {}
        if msg_list[0]["role"] == "system":
            ans["preamble"] = msg_list[0]["content"]

        role_map = {
            "user": "USER",
            "assistant": "CHATBOT",
        }
        ans["chat_history"] = []
        for msg in msg_list[1:-1]:
            ans["chat_history"].append(
                {
                    "role": role_map[msg["role"]],
                    "message": msg["content"],
                }
            )

        assert msg_list[-1]["role"] == "user"
        ans["message"] = msg_list[-1]["content"]
        return ans

    def invoke_streaming(
        self,
        prompt,
        b64image=None,
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=25,
    ):
        """
        Invokes the Cohere model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered.
            In the case of Claude 3, it is a dictionary with keys
            system -> str, messages -> list
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        tools: description of the tools that can be used, Claude format
        tool_invoker_fn: function that invokes the tools. Arguments are:
                function name - function to call
                return_results_only - we set to True because we already use Claude format
                kwargs - arguments to the tool that will be called
        :return: Inference response from the model.
        """
        # Messages that had to be added because of function use
        self.tool_use_added_msgs = []
        self.tool_results = []

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        body["stop_sequences"] = body["stop_sequences"] + extra_stop_sequences
        if len(body["stop_sequences"]) == 0:
            body.pop("stop_sequences", None)

        body["preamble"] = prompt["preamble"]
        body["chat_history"] = prompt["chat_history"].copy()
        body["message"] = prompt["message"]

        # TODO: Cohere does not support postpend. Let it use tools by itself.
        # assert postpend == "", "Model does not support postpend argument"
        postpend = ""

        # if tools is not None:
        #     body["messages"].append({"role": "assistant", "content": postpend})
        if tools is not None:
            assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

            adj_tools = []
            for x in tools:
                cur_desc = x.copy()
                required_params = cur_desc["input_schema"].copy().pop("required")
                cur_desc["parameter_definitions"] = copy.deepcopy(
                    cur_desc["input_schema"]["properties"]
                )
                cur_desc.pop("input_schema", None)
                # adjust parameter types
                lookup_param_type = {
                    "string": "str",
                    "integer": "int",
                    "boolean": "bool",
                }
                for k in cur_desc["parameter_definitions"]:
                    cur_desc["parameter_definitions"][k]["type"] = lookup_param_type[
                        cur_desc["parameter_definitions"][k]["type"]
                    ]
                # adjust required parameters
                for p in required_params:
                    cur_desc["parameter_definitions"][p]["required"] = True
                adj_tools.append(cur_desc)
            body["tools"] = adj_tools
            # assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

        cur_fail_sleep = 60
        for k in range(max_retries):
            # try:
            self.debug_body = body
            llm_body_changed = True
            while llm_body_changed:
                llm_body_changed = False
                response = self.bedrock_client.invoke_model_with_response_stream(
                    modelId=self.model_id, body=json.dumps(body)
                )
                word_count = len(
                    re.findall(r"\w+", body["message"] + str(body["chat_history"]))
                )
                print(f"Invoking {self.llm_description}. Word count: {word_count}")

                # stream responses
                partial_ans = self._response_gen(response["body"], postpend)
                x = ""
                for x in partial_ans:
                    yield x
                cur_ans = x

                if self.cur_tool_spec is not None:
                    # tool use has been required. Let's do it
                    # TODO: update upstream to reflect the inclusion of a response
                    # TODO: probably rework gradio UI to re-instantiate things every chat, or keep an instance per chat ID
                    tool_ans = tool_invoker_fn(
                        self.cur_tool_spec["tool_name"],
                        return_results_only=True,
                        **self.cur_tool_spec["input"],
                    )
                    """
                        "tool_name": func_call_spec["name"],
                        "input": func_call_spec["parameters"],
                        "generation_id": func_call_spec["generation_id"],
                    """

                    print(self.cur_tool_spec)
                    last_call = {
                        "name": self.cur_tool_spec["tool_name"],
                        "parameters": self.cur_tool_spec["input"],
                        # "generation_id": self.cur_tool_spec["generation_id"],
                    }
                    self.tool_results.append(
                        {
                            "call": last_call,
                            "outputs": [
                                {self.cur_tool_spec["tool_name"] + "_output": tool_ans}
                            ],
                        }
                    )

                    # append assistant responses
                    assistant_msg = {
                        "role": "assistant",
                        "content": [
                            {
                                "type": "text",
                                "text": cur_ans,
                            },
                            self.cur_tool_spec,
                        ],
                    }

                    next_user_msg = {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                # "generation_id": self.cur_tool_spec["generation_id"],
                                "content": tool_ans,
                            }
                        ],
                    }
                    # body["chat_history"].append(assistant_msg)
                    # body["chat_history"].append(next_user_msg)
                    body["tool_results"] = self.tool_results
                    body["chat_history"] = self.cur_tool_spec["chat_history"]
                    body["message"] = ""

                    # keep a log of messages that had to be appended due to tool use
                    self.tool_use_added_msgs.append(assistant_msg)
                    self.tool_use_added_msgs.append(next_user_msg)
                    llm_body_changed = True

                # TODO: Include proper token count and pricing
                ans_word_count = len(
                    re.findall(r"\w+", postpend + cur_ans + str(self.stop_reason))
                )
                self.word_counts.append(
                    {
                        "request_word_count": word_count,
                        "answer_word_count": ans_word_count,
                        "price_estimate": 0.001
                        * (0.003 * word_count + 0.015 * ans_word_count),
                        "exec_time_in_s": time.time() - t0,
                    }
                )
            return
        # except Exception as e:
        #    print(
        #        f'Error {str(e)}. Prompt length: {len(str(body["chat_history"]))}\n\nRetrying {k}...'
        #    )
        #    time.sleep(int(cur_fail_sleep))
        #    cur_fail_sleep *= 1.2

        raise

    def _response_gen(self, response_body, postpend=""):
        cur_ans = ""
        cur_tool_spec = None
        for x in response_body:
            out_dict = json.loads(x["chunk"]["bytes"])
            txt = ""
            if "event_type" in out_dict.keys():
                # if out_dict["event_type"] == "text-generation":
                #     txt = out_dict["text"]
                if (
                    out_dict["event_type"] == "tool-calls-generation"
                    and out_dict.get("out_dict", False)
                    and len(out_dict["tool_calls"]) > 0
                ):
                    func_call_spec = out_dict["tool_calls"][0]
                    cur_tool_spec = {
                        "tool_name": func_call_spec["name"],
                        "input": func_call_spec["parameters"],
                    }
                txt = out_dict.get("text", "")

            if txt != "":
                cur_ans += txt
                yield postpend + cur_ans
            stop_reason = out_dict["event_type"] if out_dict["is_finished"] else None
            if out_dict["is_finished"] and cur_tool_spec is not None:
                cur_tool_spec["generation_id"] = out_dict["response"]["generation_id"]
                cur_tool_spec["chat_history"] = out_dict["response"]["chat_history"]

        yield postpend + cur_ans
        # TODO: Make tool call work
        self.cur_tool_spec = cur_tool_spec
        self.stop_reason = stop_reason



import os

import numpy as np
from openai import OpenAI

rng = np.random.default_rng()


class ToolSpeechToText:
    def __init__(self):
        self.name = "speech_to_text"

        self.tool_description = {
            "name": self.name,
            "description": """Uses an automatic speech recognition tool to convert the provided audio into text in SRT format (subtitles).
Unless requested by the user, do not attempt to read the produced SRT file because the transcription may be too long.
Returns: path to a file containing a transcription of the audio in SRT format.
Raises ValueError: if not able to read the audio or video file.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "audio_file_path": {
                        "type": "string",
                        "description": "Path to the audio file that will be converted to text.",
                    },
                    "language": {
                        "type": "string",
                        "description": "Language of the audio in file audio_file_path. Should be provided in ISO-639-1 (e.g. en, pt, es) format.",
                    },
                },
                "required": ["audio_file_path", "language"],
            },
        }
        self.polly_client = None
        self.openai_client = None

    def __call__(
        self,
        audio_file_path,
        language,
        **kwargs,
    ):
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        # create client
        if self.openai_client is None:
            self.openai_client = OpenAI()

        if not os.path.isfile(audio_file_path):
            return f"Audio file not found: {audio_file_path}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/transcript_{rng_num}.srt"
        try:
            with open(audio_file_path, "rb") as audio_file:
                transcript = self.openai_client.audio.transcriptions.create(
                    model="whisper-1", response_format="srt", file=audio_file
                )
            with open(target_file, "w", encoding="UTF-8") as f:
                f.write(str(transcript))
        except Exception as e:
            return f"Transcription was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Transcription file was not saved correctly."

        ans = ["<transcript>"]
        ans.append(f"<path_to_file>{target_file}</path_to_file>")
        ans.append("</transcript>")
        return "\n".join(ans)



import os
import datetime
import importlib.resources


class RAGPromptGenerator:
    def __init__(self, use_native_tools=False):
        """Constructor.

        Arguments:

        - use_native_tools: Use LLM native tool calling abilities. In this case, we do not
            need to "teach" the LLM about tools
        """
        self.prompt = ""
        self.use_native_tools = use_native_tools
        if not use_native_tools:
            with importlib.resources.files("gat_llm.prompts").joinpath("prompt_GAT.txt").open("r", encoding="utf-8") as f:
                return f.read()

        # base prompt
        with importlib.resources.files("gat_llm.prompts").joinpath("prompt_base.txt").open("r", encoding="utf-8") as f:
            self.prompt += f.read()

        dt0 = datetime.datetime.today()
        weekday = dt0.strftime("%A")

        date_str = f"{weekday}, {dt0.year}-{str(dt0.month).zfill(2)}-{str(dt0.day).zfill(2)} in the format YYYY-MM-DD"
        date_str = f"""
<today_date>
<day_of_week>{weekday}</day_of_week>
<day>{str(dt0.day).zfill(2)}</day>
<month>{str(dt0.month).zfill(2)}</month>
<year>{dt0.year}</year>

</today_date>
        """

        self.prompt = self.prompt.replace("{{DATE}}", date_str)

        self.post_anti_hallucination = "<scratchpad> I can only use functions that have been explicitly provided. I must follow the <tool_guidelines></tool_guidelines>. I need the following tools:"



import os


class ToolReadLocalFile:
    def __init__(self):
        self.name = "read_local_files"

        self.tool_description = {
            "name": self.name,
            "description": """Reads one or more local file and return its contents. Provide one file per line.
Only read the contents of text files, like files with extensions .txt, .py, .md and others that usually contain only text.
Do not attempt to read files that are usually in binary format.

Raises ValueError: if the file does not exist.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "path_to_files": {
                        "type": "string",
                        "description": """Local path to the files whose contents should be retrieved. Provide one file per line as in the <example></example>:
<example>
file1.txt
file2.py
subfolder/file3.docx
</example>""",
                    },
                },
                "required": ["path_to_files"],
            },
        }

    def __call__(self, path_to_files, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        final_ans = ["<files>"]
        all_files = [x.strip() for x in path_to_files.splitlines() if x.strip() != ""]
        for path_to_file in all_files:
            if not os.path.isfile(path_to_file):
                ans = f"Error: Did not find file `{path_to_file}`"
                ans = f"<error>\n{ans}\n</error>"
            else:
                with open(path_to_file, "r", encoding="utf-8") as f:
                    ans = f.read()
                ans = f"<contents>\n{ans}\n</contents>"

            final_ans.append("<file>")
            final_ans.append(f"<file_name>{path_to_file}</file_name>")
            final_ans.append(ans)
            final_ans.append("</file>")

        final_ans.append("</files>")
        return "\n".join(final_ans)



import os
import json
import base64

import numpy as np
from openai import OpenAI

rng = np.random.default_rng()


class ToolTextToImage:
    def __init__(self):
        self.name = "text_to_image"

        self.tool_description = {
            "name": self.name,
            "description": """Uses a text-to-image tool to convert the provided text into an image. Uses neural methods such as diffusion.
You need to specify the text that will be converted to the image.

Raises ValueError: if not able to generate the image.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "input_text": {
                        "type": "string",
                        "description": "Text that will be converted to image.",
                    },
                },
                "required": ["input_text"],
            },
        }
        self.openai_client = None

    def __call__(
        self,
        input_text,
        **kwargs,
    ):
        os.makedirs("media", exist_ok=True)
        if self.openai_client is None:
            self.openai_client = OpenAI()
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/gen_image_{rng_num}.png"

        try:
            response = self.openai_client.images.generate(
                model="dall-e-3",
                prompt=input_text,
                n=1,
                size="1024x1024",
                response_format="b64_json",
            )
            parsed_ans = json.loads(response.json())
            revised_prompt = parsed_ans["data"][0]["revised_prompt"]
            img_content = base64.b64decode(parsed_ans["data"][0]["b64_json"])
            with open(target_file, "wb") as f:
                f.write(img_content)
        except Exception as e:
            return f"Image was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        ans = ["<image>"]
        ans.append(
            f"<revised_prompt>{revised_prompt}</revised_prompt><path_to_image>{target_file}</path_to_image>"
        )
        ans.append("</image>")
        return "\n".join(ans)



import subprocess


class ToolUseFFMPEG:
    def __init__(self):
        self.name = "use_ffmpeg"

        self.tool_description = {
            "name": self.name,
            "description": """Runs ffmpeg in the command line to manipulate videos.
If file arguments are used, specify the complete path to the file.
Save output files in the same folder as the input unless the user requests a different folder.
The command line call will be "ffmpeg [ffmpeg_arguments]".

This tool returns ffmpeg_stdout containing the stdout of ffmpeg and ffmpeg_stderr with errors returned.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "ffmpeg_arguments": {
                        "type": "string",
                        "description": """Arguments to use with ffmpeg, as in the <examples></examples>:
<examples>
<example>
<task>Extract a snippet of c:\\videos\\input.mp4 starting at 10 seconds with a duration of 20 seconds. Save the result in c:\\videos\\output.mp4</task>
<ffmpeg_arguments>-i input.mp4 -ss 10 -t 20 output.mp4</ffmpeg_arguments>
</example>
<example>
<task>Speed up input.mp4 by a factor of 2, and save the result in output.mp4</task>
<ffmpeg_arguments>-i input.mp4 -filter_complex "[0:v]setpts=0.5*PTS[v];[0:a]atempo=2.0[a]" -map "[v]" -map "[a]" output.mp4</ffmpeg_arguments>
</example>
</examples>""",
                    },
                },
                "required": ["ffmpeg_arguments"],
            },
        }

    def __call__(self, ffmpeg_arguments, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        all_args = "ffmpeg " + ffmpeg_arguments  # ffmpeg_arguments.split()
        p = subprocess.run(all_args, capture_output=True, text=True, shell=True)

        final_ans = ["<ffmpeg_stdout>"]
        final_ans.append(p.stdout + p.stderr)
        final_ans.append("</ffmpeg_stdout>")
        return "\n".join(final_ans)



# https://docs.aws.amazon.com/nova/latest/userguide/complete-request-schema.html
# https://docs.aws.amazon.com/nova/latest/userguide/tool-use-results.html
import re
import json
import time

from .base_service import LLM_Service


class LLM_Nova_Bedrock(LLM_Service):
    def __init__(self, bedrock_client, model_size):
        """Constructor
        Arguments:
            bedrock_client - Instance of boto3.client(service_name='bedrock-runtime')
                to use when making calls to bedrock models
        """
        if model_size == "Nova_Micro":
            self.model_id = "us.amazon.nova-micro-v1:0"
            self.llm_description = "Amazon Nova Micro v1.0 from AWS Bedrock (Tiny LLM)"
            self.price_per_M_input_tokens = 0.035
            self.price_per_M_output_tokens = 0.14
        elif model_size == "Nova_Lite":
            self.model_id = "us.amazon.nova-lite-v1:0"
            self.llm_description = "Amazon Nova Lite v1.0 from AWS Bedrock (Small LLM)"
            self.price_per_M_input_tokens = 0.06
            self.price_per_M_output_tokens = 0.24
        elif model_size == "Nova_Pro":
            self.model_id = "us.amazon.nova-pro-v1:0"
            self.llm_description = "Amazon Nova Pro v1.0 from AWS Bedrock (Large LLM)"
            self.price_per_M_input_tokens = 0.8
            self.price_per_M_output_tokens = 3.2

        self.bedrock_client = bedrock_client
        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "inferenceConfig": {
                "max_new_tokens": 4000,
                "temperature": 0.65,
                "stopSequences": [],
            },
        }
        # requests and answer word count
        self.word_counts = []

    def _prepare_prompt_from_list(self, msg_list):
        """Receives a list of dictionaries containing keys
        'role' and 'content' and produces the relevant formatting for the LLM
        """
        ans = {}
        if msg_list[0]["role"] == "system":
            ans["system"] = [{"text": msg_list[0]["content"]}]
        ans["messages"] = []
        for msg in msg_list[1:]:
            if isinstance(msg["content"], str):
                msg["content"] = [{"text": msg["content"]}]
            elif isinstance(msg["content"], list):
                for k in range(len(msg["content"])):
                    if msg["content"][k].get("type") == "text":
                        adj_msg = {"text": msg["content"][k]["text"]}
                        msg["content"][k] = adj_msg
                    elif msg["content"][k].get("type") == "image":
                        adj_msg = {
                            "image": {
                                "format": "jpeg",
                                "source": {
                                    "bytes": msg["content"][k]["source"]["data"],
                                },
                            }
                        }
                        msg["content"][k] = adj_msg
            ans["messages"].append(msg)
        return ans

    def invoke_streaming(
        self,
        prompt,
        b64image=None,
        postpend="",
        extra_stop_sequences=[],
        tools=None,
        tool_invoker_fn=None,
        max_retries=25,
    ):
        """
        Invokes the Nova model to run an inference
        using the input provided in the request body.

        :param prompt: The prompt to be answered.
            In the case of Claude 3, it is a dictionary with keys
            system -> str, messages -> list
        :param postpend: Extra text to append, to `put words in the mouth` of the LLM
        tools: description of the tools that can be used, Claude format
        tool_invoker_fn: function that invokes the tools. Arguments are:
                function name - function to call
                return_results_only - we set to True because we already use Claude format
                kwargs - arguments to the tool that will be called
        :return: Inference response from the model.
        """
        # Messages that had to be added because of function use
        self.tool_use_added_msgs = []

        # try:
        # The different model providers have individual request and response formats.
        # For the format, ranges, and default values for Meta Llama 2 Chat, refer to:
        # https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters-meta.html

        # start time
        t0 = time.time()
        body = self.config.copy()
        body["inferenceConfig"]["stopSequences"] = (
            body["inferenceConfig"]["stopSequences"] + extra_stop_sequences
        )

        body["messages"] = prompt["messages"].copy()

        if tools is None:
            body["messages"].append(
                {"role": "assistant", "content": [{"text": postpend}]}
            )
        else:
            adj_tools = []
            for x in tools:
                cur_desc = x.copy()
                cur_desc["inputSchema"] = {"json": cur_desc["input_schema"]}
                cur_desc.pop("input_schema", None)
                adj_tools.append({"toolSpec": cur_desc})
            body["toolConfig"] = {"tools": adj_tools}
            assert postpend == "", "When using tools, postpend is not supported"
            assert (
                tool_invoker_fn is not None
            ), "When using tools, a tool invoker must be provided"

        cur_fail_sleep = 6
        for k in range(max_retries):
            try:
                llm_body_changed = True
                while llm_body_changed:
                    llm_body_changed = False
                    response = self.bedrock_client.invoke_model_with_response_stream(
                        modelId=self.model_id, body=json.dumps(body)
                    )
                    word_count = len(re.findall(r"\w+", str(body["messages"])))
                    print(f"Invoking {self.llm_description}. Word count: {word_count}")

                    # stream responses
                    # return response
                    partial_ans = self._response_gen(response["body"], postpend)
                    for x in partial_ans:
                        yield x
                    cur_ans = x

                    if self.cur_tool_spec is not None:
                        # tool use has been required. Let's do it
                        # TODO: update upstream to reflect the inclusion of a response
                        # TODO: probably rework gradio UI to re-instantiate things every chat, or keep an instance per chat ID
                        tool_ans = tool_invoker_fn(
                            self.cur_tool_spec["toolUse"]["name"],
                            return_results_only=True,
                            **self.cur_tool_spec["toolUse"]["input"],
                        )

                        # append assistant responses
                        assistant_msg = {
                            "role": "assistant",
                            "content": [
                                {
                                    "text": cur_ans,
                                },
                            ],
                        }
                        assistant_msg2 = {
                            "role": "assistant",
                            "content": [
                                {
                                    "toolUse": self.cur_tool_spec["toolUse"],
                                },
                            ],
                        }

                        next_user_msg = {
                            "role": "user",
                            "content": [
                                {
                                    "toolResult": {
                                        # "type": "tool_result",
                                        "toolUseId": self.cur_tool_spec["toolUse"][
                                            "toolUseId"
                                        ],
                                        "content": [{"text": tool_ans}],
                                    }
                                }
                            ],
                        }

                        body["messages"].append(assistant_msg)
                        body["messages"].append(assistant_msg2)
                        body["messages"].append(next_user_msg)

                        # keep a log of messages that had to be appended due to tool use
                        self.tool_use_added_msgs.append(assistant_msg)
                        self.tool_use_added_msgs.append(assistant_msg2)
                        self.tool_use_added_msgs.append(next_user_msg)
                        llm_body_changed = True

                    # TODO: Include proper token count and pricing
                    ans_word_count = len(
                        re.findall(r"\w+", postpend + cur_ans + str(self.stop_reason))
                    )
                    self.word_counts.append(
                        {
                            "request_word_count": word_count,
                            "answer_word_count": ans_word_count,
                            "price_estimate": 0.001
                            * (0.003 * word_count + 0.015 * ans_word_count),
                            "exec_time_in_s": time.time() - t0,
                        }
                    )
                    # """
                return

            except Exception as e:
                print(
                    f'Error {str(e)}. Prompt length: {len(str(body["messages"]))}\n\nRetrying {k}...'
                )
                time.sleep(int(cur_fail_sleep))
                cur_fail_sleep *= 1.2

        raise
        # return response

        # except ClientError:
        #    logger.error("Couldn't invoke Claude")
        #    raise

    def _response_gen(self, response_body, postpend=""):
        cur_ans = ""
        cur_tool_spec = None
        for x in response_body:
            out_dict = json.loads(x["chunk"]["bytes"])
            txt = ""
            if "contentBlockDelta" in out_dict.keys():
                if (
                    out_dict["contentBlockDelta"]["delta"].get("text")
                    and cur_tool_spec is None
                ):
                    txt = out_dict["contentBlockDelta"]["delta"]["text"]
                elif out_dict["contentBlockDelta"]["delta"].get("toolUse"):
                    cur_tool_spec["toolUse"]["input"] = out_dict["contentBlockDelta"][
                        "delta"
                    ]["toolUse"]["input"]

            if "contentBlockStart" in out_dict.keys():
                if out_dict["contentBlockStart"]["start"].get("toolUse"):
                    cur_tool_spec = out_dict["contentBlockStart"]["start"].copy()

            if cur_tool_spec is None:
                cur_ans += txt
                yield postpend + cur_ans

            stop_reason = None
            if out_dict.get("messageStop"):
                stop_reason = out_dict["messageStop"].get("stopReason", None)
            if stop_reason is not None:
                break
        if cur_tool_spec is not None:
            cur_tool_spec["toolUse"]["input"] = json.loads(
                cur_tool_spec["toolUse"]["input"]
            )
        self.cur_tool_spec = cur_tool_spec
        self.stop_reason = stop_reason



import datetime
import dateutil


class ToolDoDateMath:
    def __init__(self):
        self.name = "do_date_math"

        self.tool_summary = f"""<tool_summary>
<tool_name>{self.name}</tool_name>
<summary>Adds or subtracts one or more time intervals from a given date in the format YYYY-MM-DD.</summary>
</tool_summary>"""

        self.tool_description = {
            "name": self.name,
            "description": """Adds or subtracts one or more time intervals from a given date in the format YYYY-MM-DD.

The <deltas></deltas> to be added or subtracted should be separated by commas. Use negative values to subtract, as shown in the <example_deltas></example_deltas>:

<example_deltas>
<example_delta>5<example_delta>
<example_delta>-7, -14, -21<example_delta>
<example_delta>5, -6, -8<example_delta>
</example_deltas>

Raises ValueError: if one of the parameters is invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "base_date": {
                        "type": "string",
                        "description": "Base date in the format YYYY-MM-DD",
                    },
                    "deltas": {
                        "type": "string",
                        "description": "Intervals, as defined in delta_type, to add or subtract from the base date, separated by commas",
                    },
                    "delta_type": {
                        "type": "string",
                        "enum": ["day", "week", "month", "year"],
                        "description": "Type of interval to sum or subtract from base_date. Possible values are listed in the string enum: ['day', 'week', 'month', 'year']",
                    },
                },
                "required": ["base_date", "deltas", "delta_type"],
            },
        }

        self.tool_summary = self.tool_description

    def __call__(self, base_date, deltas, delta_type, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        allowed_delta_types = ["day", "week", "month", "year"]
        if delta_type not in allowed_delta_types:
            return f"Error: delta_type must be one of {allowed_delta_types}"

        try:
            date_object = datetime.datetime.strptime(base_date, "%Y-%m-%d").date()
        except:
            return "Error: please provide a base_date in the format YYYY-MM-DD"

        delta_periods = [int(x.strip()) for x in deltas.split(",")]

        # dateutil.relativedelta.relativedelta(dt1=None, dt2=None, years=0, months=0, days=0, leapdays=0, weeks=0

        if delta_type == "day":
            final_deltas = [
                dateutil.relativedelta.relativedelta(days=x) for x in delta_periods
            ]
        elif delta_type == "week":
            final_deltas = [
                dateutil.relativedelta.relativedelta(weeks=x) for x in delta_periods
            ]
        elif delta_type == "month":
            final_deltas = [
                dateutil.relativedelta.relativedelta(months=x) for x in delta_periods
            ]
        elif delta_type == "year":
            final_deltas = [
                dateutil.relativedelta.relativedelta(years=x) for x in delta_periods
            ]

        ans = [(date_object + x).strftime("%Y-%m-%d %A") for x in final_deltas]
        return ",".join(ans)
